﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.BinPacking.Decoders;
using HeuristicLab.Problems.BinPacking.Interfaces;
using HeuristicLab.Problems.BinPacking.PackingBin;
using HeuristicLab.Problems.BinPacking.PackingItem;

namespace HeuristicLab.Encodings.PackingEncoding.MultiComponentVector {
  [Item("MultiComponentVectorMoveGenerator", "Base class for all multi-component vector move generators.")]
  [StorableClass]
  public abstract class MultiComponentVectorMoveGenerator : SingleSuccessorOperator, IMultiComponentVectorMoveOperator, IMoveGenerator {
    public override bool CanChangeName {
      get { return false; }
    }
    public ILookupParameter<MultiComponentVectorEncoding> MultiComponentVectorParameter {
      get { return (ILookupParameter<MultiComponentVectorEncoding>)Parameters["MultiComponentVector"]; }
    }
    public ILookupParameter<IPackingMove> PackingMoveParameter {
      get { return (LookupParameter<IPackingMove>)Parameters["PackingMove"]; }
    }
    protected ScopeParameter CurrentScopeParameter {
      get { return (ScopeParameter)Parameters["CurrentScope"]; }
    }               


    [StorableConstructor]
    protected MultiComponentVectorMoveGenerator(bool deserializing) : base(deserializing) { }
    protected MultiComponentVectorMoveGenerator(MultiComponentVectorMoveGenerator original, Cloner cloner) : base(original, cloner) { }
    public MultiComponentVectorMoveGenerator()
      : base() {
        Parameters.Add(new LookupParameter<MultiComponentVectorEncoding>("MultiComponentVector", "The multi component vector for which moves should be generated."));
        Parameters.Add(new LookupParameter<IPackingMove>("PackingMove", "The moves that should be generated in subscopes."));
      Parameters.Add(new ScopeParameter("CurrentScope", "The current scope where the moves should be added as subscopes."));
    }


    public static int NrOfRotationMoves(MultiComponentVectorEncoding mcv) {
      return mcv.NrOfItems;
    }
    public static IEnumerable<MultiComponentVectorMove> GenerateRotationMoves(MultiComponentVectorEncoding multiComponentVector) {
      for (int binNr = 0; binNr < multiComponentVector.PackingInformations.Count; binNr++)
        for (int index = 0; index < multiComponentVector.PackingInformations[binNr].Count; index++)
          yield return new SingleItemRotationMove(binNr, index, multiComponentVector);
    }

    public static int NrOfSingleGroupingMoves(MultiComponentVectorEncoding mcv) {
      return mcv.NrOfItems * (mcv.NrOfBins - 1);
    }
    public static IEnumerable<MultiComponentVectorMove> GenerateSingleGroupingMoves(MultiComponentVectorEncoding multiComponentVector) {
      for (int oldBin = 0; oldBin < multiComponentVector.PackingInformations.Count; oldBin++)
        for (int index = 0; index < multiComponentVector.PackingInformations[oldBin].Count; index++)
          for (int newBin = 0; newBin < multiComponentVector.NrOfBins; newBin++) 
            if (newBin != oldBin)
              yield return new SingleGroupingMove(oldBin, index, newBin, multiComponentVector);
    }

    public static int NrOfSwapPositionMoves(MultiComponentVectorEncoding mcv) {
      int nrOfSwapPositionMoves = 0;
      foreach (var groupInformations in mcv.PackingInformations) {
        nrOfSwapPositionMoves += (groupInformations.Value.Count * (groupInformations.Value.Count - 1)) / 2;
      }
      return nrOfSwapPositionMoves;
    }
    public static IEnumerable<MultiComponentVectorMove> GenerateSwapPositionMoves(MultiComponentVectorEncoding multiComponentVector) {
      for (int bin = 0; bin < multiComponentVector.PackingInformations.Count; bin++)
        for (int oldIndex = 0; oldIndex < multiComponentVector.PackingInformations[bin].Count - 1; oldIndex++) 
          for (int newIndex = oldIndex + 1; newIndex < multiComponentVector.PackingInformations[bin].Count; newIndex++)
            yield return new SwapPositionMove(bin, oldIndex, newIndex, multiComponentVector);
    }

    public static int NrOfChangePositionMoves(MultiComponentVectorEncoding mcv) {
      int nrOfChangePositionMoves = 0;
      foreach (var groupInformations in mcv.PackingInformations) {
        nrOfChangePositionMoves += (groupInformations.Value.Count * (groupInformations.Value.Count - 1)) / 2;
      }
      return nrOfChangePositionMoves;
    }
    public static IEnumerable<MultiComponentVectorMove> GenerateChangePositionMoves(MultiComponentVectorEncoding multiComponentVector) {
      for (int bin = 0; bin < multiComponentVector.PackingInformations.Count; bin++)
        for (int oldIndex = 0; oldIndex < multiComponentVector.PackingInformations[bin].Count - 1; oldIndex++) 
          for (int newIndex = oldIndex + 1; newIndex < multiComponentVector.PackingInformations[bin].Count; newIndex++)
            yield return new ChangePositionMove(bin, oldIndex, newIndex, multiComponentVector);
    }

    public override IOperation Apply() {
      MultiComponentVectorEncoding mcv = MultiComponentVectorParameter.ActualValue;
      MultiComponentVectorMove[] moves = GenerateMoves(mcv);
      Scope[] moveScopes = new Scope[moves.Length];
      for (int i = 0; i < moveScopes.Length; i++) {
        moveScopes[i] = new Scope(i.ToString());
        moveScopes[i].Variables.Add(new Variable(PackingMoveParameter.ActualName, moves[i]));
      }
      CurrentScopeParameter.ActualValue.SubScopes.AddRange(moveScopes);
      return base.Apply();
    }

    protected abstract MultiComponentVectorMove[] GenerateMoves(MultiComponentVectorEncoding multiComponentVector);
  }
}
